@@ -4,7 +4,7 @@ use Mojo::Base 'Mojolicious::Plugin';
 use JSON::Validator::OpenAPI;
 use constant DEBUG => $ENV{MOJO_OPENAPI_DEBUG} || 0;
 
-our $VERSION = '0.01';
+our $VERSION = '0.02';
 
 my $EXCEPTION = {errors => [{message => 'Internal server error.', path => '/'}], status => 500};
 my $NOT_IMPLEMENTED = {errors => [{message => 'Not implemented.', path => '/'}], status => 501};
@@ -22,7 +22,7 @@ sub register {
   $app->helper('reply.openapi'    => \&_reply);
   $app->hook(before_render => \&_before_render);
 
-  $self->{log_level} = $config->{log_level} || 'warn';    # TODO: Is warn a nice default?
+  $self->{log_level} = $ENV{MOJO_OPENAPI_LOG_LEVEL} || $config->{log_level} || 'warn';
   $self->_validator->schema($api_spec->data)->coerce($config->{coerce} // 1);
   $self->_add_routes($app, $api_spec, $config->{route});
 }
@@ -73,12 +73,18 @@ sub _add_routes {
 
 sub _before_render {
   my ($c, $args) = @_;
-  return if !$args->{exception} and grep {/^\w+$/} keys %$args;    # TODO: Is this robust?
+
+  # TODO: Is this robust enough?
+  # Want to disable this hook if the user does $c->render(template => "foo.html");
+  return if !$args->{exception} and grep {/^\w+$/} keys %$args;
   return unless $c->stash('openapi.op_spec');
   my $format = $c->stash('format') || 'json';
   my $io = $args->{exception} ? $EXCEPTION : $c->stash('openapi.io') || $NOT_IMPLEMENTED;
-  $args->{status}  = delete $io->{status};
-  $args->{$format} = $io;                                          # TODO: Is $format good enough?
+  $args->{status} = delete $io->{status};
+
+  # TODO: Is $format a good idea? Was thinking someone might want to set
+  # $c->stash(format => "xml") and it "should just work"
+  $args->{$format} = $io;
 }
 
 sub _input {
@@ -141,7 +147,6 @@ sub _validate {
       $c->stash('openapi.io' => {errors => \@errors, status => 500});
       $self->_log($c, '>>>', \@errors);
     }
-    warn "[OpenAPI] >>> @{[$c->req->url]} == (@errors)\n" if DEBUG;
   }
   else {
     @errors = $self->_validator->validate_request($c, $op_spec, \my %input);
@@ -152,7 +157,6 @@ sub _validate {
     else {
       $c->stash('openapi.input' => \%input);
     }
-    warn "[OpenAPI] <<< @{[$c->req->url]} == (@errors)\n" if DEBUG;
   }
 
   return @errors;
@@ -170,30 +174,37 @@ Mojolicious::Plugin::OpenAPI - OpenAPI / Swagger plugin for Mojolicious
 
 =head2 Specification
 
+This plugin reads an L<OpenAPI specification|https://openapis.org/specification>
+and generate routes and input/output rules from it. See L<JSON::Validator> for
+L<supported schema formats|JSON::Validator/Supported schema formats>.
+
   {
+    "basePath": "/api",
     "paths": {
       "/pets": {
         "get": {
           "x-mojo-to": "pet#list",
           "summary": "Finds pets in the system",
+          "parameters": [],
           "responses": {
-          "200": {
-            "description": "Pet response",
-            "schema": { "type": "array", "items": { "$ref": "#/definitions/Pet" } }
-          },
-          "default": {
-            "description": "Unexpected error",
-            "schema": { "$ref": "http://git.io/vcKD4#" }
+            "200": {
+              "description": "Pet response",
+              "schema": { "type": "array", "items": { "type": "object" } }
+            },
+            "default": {
+              "description": "Unexpected error",
+              "schema": { "$ref": "http://git.io/vcKD4#" }
+            }
           }
         }
       }
     }
   }
 
-The important part in the spec above is "x-mojo-to". The "x-mojo-to" key can
+The non-standard part in the spec above is "x-mojo-to". The "x-mojo-to" key can
 either a plain string, object (hash) or an array. The string and hash will be
 passed directly to L<Mojolicious::Routes::Route/to>, while the array ref, will
-be flattened first.
+be flattened first. Examples:
 
   "x-mojo-to": "pet#list"
   $route->to("pet#list");
@@ -204,18 +215,27 @@ be flattened first.
   "x-mojo-to": ["pet#list", {"foo": 123}]
   $route->to("pet#list", {foo => 123});
 
+The complete HTTP request for getting the "pet list" will be C<GET /api/pets>
+The first part of the path ("/api") comes from C<basePath>, the second part
+comes from the key under C<paths>, and the HTTP method comes from the key under
+C</pets>.
+
+C<parameters> and C<responses> will be used to define rules for
+L<input|/openapi.input> and L<output|reply.openapi>.
+
 =head2 Application
 
   package Myapp;
   use Mojolicious;
 
-  sub register {
+  sub startup {
     my $app = shift;
-    $app->plugin("OpenAPI" => {url => "myapi.json"});
+    $app->plugin("OpenAPI" => {url => $app->home->rel_file("myapi.json")});
   }
 
-See L</register> for information about what the plugin config can be, in
-addition to "url".
+The first thing in your code that you need to do is to load this plugin and the
+L</Specification>. See L</register> for information about what the plugin
+config can be, in addition to "url".
 
 =head2 Controller
 
@@ -238,8 +258,11 @@ addition to "url".
     $c->reply->openapi($output, 200);
   }
 
-The controller input and output will only be validated if the L</openapi.input>
-and L</reply.openapi> methods are used.
+The input and output to the action will only be validated if the
+L</openapi.input> and L</reply.openapi> methods are used.
+
+All OpenAPI powered actions will have L<auto-rendering|/reply.openapi> enabled,
+which means that the C<return;> above will render an error document.
 
 =head1 DESCRIPTION
 
@@ -253,27 +276,25 @@ This plugin is currently EXPERIMENTAL.
 
   $hash = $c->openapi->input;
 
-Returns the data which has been L<validated|/openapi.validate> by the in
-OpenAPI specification.
+Returns the request parameters if they are L<valid|/openapi.validate>, and
+C<undef> on invalid input.
 
 =head2 openapi.spec
 
   $hash = $c->openapi->spec;
 
-Returns the OpenAPI specification for the current route:
+Returns the OpenAPI specification for the current route. Example:
 
   {
     "paths": {
       "/pets": {
         "get": {
-          // This datastructure
+          // This datastructure is returned
         }
       }
     }
   }
 
-Note: This might return a JSON pointer in the future.
-
 =head2 openapi.validate
 
   # validate request
@@ -287,12 +308,15 @@ L</openapi.input>.
 
 =head2 reply.openapi
 
-  $c->reply->openapi($output, $http_status);
-  $c->reply->openapi;
+  $c->reply->openapi(\%output, $http_status);
 
-Will L<validate|/openapi.validate> C<$output> before passing it on to
-L<Mojolicious::Controller/render>. Calling this helper without any arguments
-will cause auto-rendering of input errors. See L</SYNOPSIS> for example.
+Will L<validate|/openapi.validate> C<%output> before passing it on to
+L<Mojolicious::Controller/render>. Note that C<%output> will be passed on using
+the L<format|Mojolicious::Guides::Rendering/Content type> key in stash, which
+defaults to "json". This also goes for L<auto-rendering|/Controller>. Example:
+
+  my $format = $c->stash("format") || "json";
+  $c->render($format => \%output);
 
 =head1 METHODS
 
@@ -300,30 +324,53 @@ will cause auto-rendering of input errors. See L</SYNOPSIS> for example.
 
   $self->register($app, \%config);
 
-Loads the OpenAPI specification, validates it and add routes to L<$app>.
-It will also set up L</HELPERS>. C<%config> can have:
+Loads the OpenAPI specification, validates it and add routes to
+L<$app|Mojolicious>. It will also set up L</HELPERS> and adds a
+L<before_render|Mojolicious/before_render> hook for auto-rendering of error
+documents.
 
-  {
-    coerce    => 0,                           # default: 1
-    log_level => "debug",                     # default: warn
-    route     => $app->routes->under(...)     # not required
-    url       => "path/to/specification.json" # required
-  }
-
-C<route> can be specified in case you want to have a protected API.
-
-See L<JSON::Validator/coerce> for possible values that C<coerce> can take.
-
-See L<JSON::Validator/schema> for the different C<url> formats that is
-accepted. Note that relative paths will be relative to L<Mojo/home>.
-
-=head1 TODO
+C<%config> can have:
 
 =over 2
 
-=item * Add WebSockets support.
+=item * coerce
 
-=item * Add support for /api.html (human readable format)
+See L<JSON::Validator/coerce> for possible values that C<coerce> can take.
+
+Default: 1
+
+=item * log_level
+
+C<log_level> is used when logging invalid request/response error messages.
+
+Default: "warn".
+
+=item * route
+
+C<route> can be specified in case you want to have a protected API. Example:
+
+  $app->plugin(OpenAPI => {
+    route => $app->routes->under("/api")->to("user#auth"),
+    url   => $app->home->rel_file("cool.api"),
+  });
+
+=item * url
+
+See L<JSON::Validator/schema> for the different C<url> formats that is
+accepted.
+
+=back
+
+=head1 TODO
+
+This plugin is still a big rough on the edges, but I decided to release it on
+CPAN so people can start playing around with it.
+
+=over 2
+
+=item * Add L<WebSockets support|https://github.com/jhthorsen/mojolicious-plugin-openapi/compare/websocket>.
+
+=item * Add support for /api.html (human readable documentation)
 
 =item * Never add support for "x-mojo-around-action", but possibly "before action".
 
